#include <iostream>
#include <mutex>
#include <condition_variable>
#include <thread>
#include <mpirxx.h>

#include "Constants.h"
#include "Utility.h"
#include "Timer.h"

using namespace std;
using namespace std::chrono;

/**
 * Shared data provided to each thread.
 */
struct ThreadData
{
    mutex *powers2_mutex;
    mutex *y_mutex;
    condition_variable *next_2_done;
    condition_variable *next_y_done;
    int *index_2;
    int *index_y;

    // Values of the powers of 2 computed by the powers2 thread.
    mpf_class *powers2;

    // Values of y computed by the y thread.
    mpf_class *y;
};

/**
 * Thread to compute consecutive powers of 2.
 * @param tdata the shared thread data.
 */
void compute_powers2(ThreadData tdata)
{
    const mpf_class FOUR(4);

    mpf_class *powers2 = tdata.powers2;
    int& index_2 = *tdata.index_2;
    condition_variable& next_2_done = *tdata.next_2_done;

    powers2[0] = 2;

    for (int i = 1; i <= MAX_ITERATIONS; i++)
    {
        powers2[i] = FOUR*powers2[i-1];

        // Signal the next power of 2 is ready.
        index_2 = i;
        next_2_done.notify_one();

        this_thread::yield();
    }
}

/**
 * Thread to compute consecutive values of y.
 * @param tdata the shared thread data.
 */
void compute_ys(ThreadData tdata)
{
    const mpf_class ONE(1), TWO(2), FOUR(4), SIX(6);
    const mpf_class SQRT2 = sqrt(TWO);

    mpf_class *y = tdata.y;
    int& index_y = *tdata.index_y;
    condition_variable& next_y_done = *tdata.next_y_done;

    mpf_class y4, yRoot4;

    y[0] = SQRT2 - ONE;

    for (int i = 1; i <= MAX_ITERATIONS; i++)
    {
        y4 = power4(y4, y[i-1]);
        yRoot4 = root4(yRoot4, ONE - y4);
        y[i] = (ONE - yRoot4)/(ONE + yRoot4);

        // Signal the next y is ready.
        index_y = i;
        next_y_done.notify_one();

        this_thread::yield();
    }
}

/**
 * Compute pi in multiple threads.
 * @param pi to be computed.
 */
void mtt_compute_pi(mpf_class& pi)
{
    // Shared thread data;
    mutex powers2_mutex;
    mutex y_mutex;
    condition_variable next_2_done;
    condition_variable next_y_done;
    int index_2 = -1;
    int index_y = -1;

    // Values of the powers of 2 computed by the powers2 thread.
    mpf_class powers2[MAX_ITERATIONS + 1];

    // Values of y computed by the y thread.
    mpf_class y[MAX_ITERATIONS + 1];

    ThreadData tdata { &powers2_mutex, &powers2_mutex,
                       &next_2_done, &next_y_done,
                       &index_2, &index_y, powers2, y };

    const mpf_class ONE(1), TWO(2), FOUR(4), SIX(6);
    const mpf_class SQRT2 = sqrt(TWO);

    mpf_class a, y4, yRoot4, aTerm;
    mpf_class a_prev = SIX - FOUR*SQRT2;

    // Start the powers2 thread.
    thread powers2_thread = thread(compute_powers2, tdata);
    this_thread::yield();

    // Start the y thread;
    thread y_thread = thread(compute_ys, tdata);
    this_thread::yield();

    for (int i = 1; i <= MAX_ITERATIONS; i++)
    {
        // Wait for the next y to be done.
        unique_lock<mutex> lock_y(y_mutex);
        next_y_done.wait(lock_y, [&i, &index_y] { return i <= index_y; });
        lock_y.unlock();

        mpf_class temp;
        aTerm = a_prev*power4(temp, ONE + y[i]);

        // Wait for the next power of 2 to be done.
        unique_lock<mutex> lock_2(powers2_mutex);
        next_2_done.wait(lock_2, [&i, &index_2] { return i <= index_2; });
        lock_2.unlock();

        a = aTerm - powers2[i]*y[i]*(ONE + y[i] + y[i]*y[i]);
        a_prev = a;
    }

    pi = ONE/a;

    // Wait for the threads to complete.
    powers2_thread.join();
    y_thread.join();
}

void mtt_compute_and_print_pi()
{
    mpf_set_default_prec(BIT_COUNT*PRECISION/2);

    cout << "Multithreaded calculation of pi to "
         << commafy(PLACES) << " places:" << endl;

    // Time the computation of pi.
    mpf_class pi;
    long ns = Timer::time_it(mtt_compute_pi, pi);

    // Convert the number to a string for printing.
    mp_exp_t exp;      // exponent (not used)
    char *str = NULL;
    char *s = mpf_get_str(str, &exp, BASE, 0, pi.get_mpf_t());

    print(s, ns);
}
